;; Generated package description from gle-mode.el  -*- no-byte-compile: t -*-
(define-package "gle-mode" "1.1.0.20221108.145352" "Major mode to edit Graphics Layout Engine files" '((cl-lib "0.5")) :commit "4b850659cd23694a5fd80ef7d1b26da27bebd246" :url "https://elpa.gnu.org/packages/gle-mode.html" :authors '(("Stefan Monnier" . "monnier@iro.umontreal.ca")) :maintainer '("Stefan Monnier" . "monnier@iro.umontreal.ca"))
